# coding=utf-8
"""
Bootstrap module parses command line arguments and initializes the app.

Configures the database connection and starts the web application.
"""
import locale
import sys

import click
from flask import Flask
from gevent.pywsgi import WSGIServer

from sacredboard.app.config import jinja_filters
from sacredboard.app.data.filestorage import FileStorage
from sacredboard.app.data.mongodb import PyMongoDataAccess
from sacredboard.app.webapi import routes, metrics

locale.setlocale(locale.LC_ALL, '')
app = Flask(__name__)


@click.command()
@click.option("-m", default=None, metavar="HOST:PORT:DATABASE",
              help="Connect to MongoDB using the format"
                   " host:port:database_name or just the database_name. "
                   "Default: sacred"
                   " Mutually exclusive with -mu")
@click.option("-mu", default=(None, None),
              metavar="CONNECTION_STRING DATABASE", type=(str, str),
              help="Connect to MongoDB using mongodb://..."
                   " and specify the database name."
                   " Mutually exclusive with -m")
@click.option("-mc", default="runs", metavar="COLLECTION",
              help="The collection containing the Sacred's list of runs. "
                   "You might need it if you use a custom collection name "
                   "or Sacred v0.6 (which used default.runs). "
                   "Default: runs")
@click.option("-F", default="",
              help="Path to directory containing experiment results of the"
                   "File Storage observer. (experimental)")
@click.option("--no-browser", is_flag=True, default=False,
              help="Do not open web browser automatically.")
@click.option("--debug", is_flag=True, default=False,
              help="Run the application in Flask debug mode "
                   "(for development).")
@click.version_option()
def run(debug, no_browser, m, mu, mc, f):
    """
    Sacredboard.

\b
Sacredboard is a monitoring dashboard for Sacred.
Homepage: http://github.com/chovanecm/sacredboard

Example usage:

\b
sacredboard -m sacred
    Starts Sacredboard on default port (5000) and connects to
    a local MongoDB database called 'sacred'. Opens web browser.
    Note: MongoDB must be listening on localhost.
\b
sacredboard -m 192.168.1.1:27017:sacred
    Starts Sacredboard on default port (5000) and connects to
    a MongoDB database running on 192.168.1.1 on port 27017
    to a database called 'sacred'. Opens web browser.
\b
sacredboard -mu mongodb://user:pwd@host/admin?authMechanism=SCRAM-SHA-1 sacred
    Starts Sacredboard on default port (5000) and connects to
    a MongoDB database running on localhost on port 27017
    to a database called 'sacred'. Opens web browser.

\b
sacredboard -m sacred -mc default.runs
    Starts Sacredboard on default port (5000) and connects to
    a local MongoDB database called 'sacred' and uses the Sacred's 0.6
    default collection 'default.runs' to search the runs in.
    Opens web browser.
    Note: MongoDB must be listening on localhost.

    """
    if m or mu != (None, None):
        add_mongo_config(app, m, mu, mc)
        app.config["data"].connect()
    elif f:
        app.config["data"] = FileStorage(f)
    else:
        print("Must specify either a mongodb instance or " +
              "a path to a file storage.\nRun sacredboard --help "
              "for more information.", file=sys.stderr)
        sys.exit(1)

    app.config['DEBUG'] = debug
    app.debug = debug
    jinja_filters.setup_filters(app)
    routes.setup_routes(app)
    metrics.initialize(app)

    if debug:
        app.run(host="0.0.0.0", debug=True)
    else:
        for port in range(5000, 5050):
            http_server = WSGIServer(('0.0.0.0', port), app)
            try:
                http_server.start()
            except OSError as e:
                # try next port
                continue
            print("Starting sacredboard on port %d" % port)
            if not no_browser:
                click.launch("http://127.0.0.1:%d" % port)
            http_server.serve_forever()
            break


def add_mongo_config(app, simple_connection_string,
                     mongo_uri, collection_name):
    """
    Configure the application to use MongoDB.

    :param app: Flask application
    :param simple_connection_string:
                Expects host:port:database_name or database_name
                Mutally_exclusive with mongo_uri
    :param mongo_uri: Expects mongodb://... as defined
                in https://docs.mongodb.com/manual/reference/connection-string/
                Mutually exclusive with simple_connection_string (must be None)
    :param collection_name: The collection containing Sacred's runs
    :return:
    """
    if mongo_uri != (None, None):
        add_mongo_config_with_uri(app, mongo_uri[0], mongo_uri[1],
                                  collection_name)
        if simple_connection_string is not None:
            print("Ignoring the -m option. Overridden by "
                  "a more specific option (-mu).", file=sys.stderr)
    else:
        # Use the default value 'sacred' when not specified
        if simple_connection_string is None:
            simple_connection_string = "sacred"
        add_mongo_config_simple(app, simple_connection_string, collection_name)


def add_mongo_config_simple(app, connection_string, collection_name):
    """
    Configure the app to use MongoDB.

    :param app: Flask Application
    :type app: Flask
    :param connection_string: in format host:port:database or database
            (default: sacred)
    :type connection_string: str

    :param collection_name: Name of the collection
    :type collection_name: str
    """
    split_string = connection_string.split(":")
    config = {"host": "localhost", "port": 27017, "db": "sacred"}
    if len(split_string) > 0 and len(split_string[-1]) > 0:
        config["db"] = split_string[-1]
    if len(split_string) > 1:
        config["port"] = int(split_string[-2])
    if len(split_string) > 2:
        config["host"] = split_string[-3]
    app.config["data"] = PyMongoDataAccess.build_data_access(
        config["host"], config["port"], config["db"], collection_name)


def add_mongo_config_with_uri(app, connection_string_uri,
                              database_name, collection_name):
    """
    Configure PyMongo with a MongoDB connection string.

    :param app: Flask application
    :param connection_string_uri: MongoDB connection string
    :param database_name: Sacred database name
    :param collection_name: Sacred's collection with runs
    :return:
    """
    app.config["data"] = PyMongoDataAccess.build_data_access_with_uri(
        connection_string_uri, database_name, collection_name
    )


if __name__ == '__main__':
    run()



# coding=utf-8
"""
Configures filters for use in pages rendered by Flask.

It helps writing the frontend templates,
 e.g. a data variable can be rendered in local format using
 {{date_variable | format_datetime}}
"""
import datetime

import simplejson
from bson import json_util
from flask import Blueprint

filters = Blueprint("filters", __name__)


@filters.app_template_filter("format_datetime")
def format_datetime(value):
    """Format datetime according to server's locale."""
    return value.strftime('%X %x')


@filters.app_template_filter("timediff")
def timediff(time):
    """Return the difference in seconds between now and the given time."""
    now = datetime.datetime.utcnow()
    diff = now - time
    diff_sec = diff.total_seconds()
    return diff_sec


@filters.app_template_filter("last_line")
def last_line(text):
    """
    Get the last meaningful line of the text, that is the last non-empty line.

    :param text: Text to search the last line
    :type text: str
    :return:
    :rtype: str
    """
    last_line_of_text = ""
    while last_line_of_text == "" and len(text) > 0:
        last_line_start = text.rfind("\n")
        # Handle one-line strings (without \n)
        last_line_start = max(0, last_line_start)
        last_line_of_text = text[last_line_start:].strip("\r\n ")
        text = text[:last_line_start]
    return last_line_of_text


@filters.app_template_filter("first_letter")
def first_letter(text):
    """Return the first leter of the text."""
    return text[:1]


@filters.app_template_filter("dump_json")
def dump_json(obj):
    """Dump Python object as JSON string."""
    return simplejson.dumps(obj, ignore_nan=True, default=json_util.default)


@filters.app_template_filter("tostr")
def tostr(obj):
    """Convert object to string."""
    return str(obj)


@filters.app_template_filter("detect_alive_experiment")
def detect_alive_experiment(time_difference):
    """Decide whether experiment is alive or not."""
    return time_difference < 120


def setup_filters(app):
    """Register the filters in a Flask application."""
    app.register_blueprint(filters)



"""
Interface for accessing Sacred metrics.

Issue: https://github.com/chovanecm/sacredboard/issues/60
"""


class MetricsDAO:
    """
    Interface for accessing Sacred metrics.

    Issue: https://github.com/chovanecm/sacredboard/issues/58
    """

    def get_metric(self, run_id, metric_id):
        """
        Read a metric of the given id and run.

        The returned object has the following format (timestamps are datetime
         objects).

        .. code::

            {"steps": [0,1,20,40,...],
            "timestamps": [timestamp1,timestamp2,timestamp3,...],
            "values": [0,1 2,3,4,5,6,...],
            "name": "name of the metric",
            "metric_id": "metric_id",
            "run_id": "run_id"}

        :param run_id: ID of the Run that the metric belongs to.
        :param metric_id: The ID fo the metric.
        :return: The whole metric as specified.

        :raise NotFoundError
        """
        raise NotImplementedError("The MetricsDAO method is abstract.")



"""
Generic DAO object for safe access to the MongoDB.

Issue: https://github.com/chovanecm/sacredboard/issues/61
"""
import pymongo
from pymongo.errors import InvalidName

from sacredboard.app.data import DataSourceError
from .mongocursor import MongoDbCursor


class GenericDAO:
    """
    Generic DAO object for safe access to the MongoDB.

    Issue: https://github.com/chovanecm/sacredboard/issues/61
    """

    def __init__(self, pymongo_client, database_name):
        """
        Create a new GenericDAO object that will work on the given database.

        :param pymongo_client: PyMongo client that is connected to MongoDB.
        :param database_name: Name of the database this GenericDAO works with.

        :raise DataSourceError
        """
        self._client = pymongo_client
        self._database = self._get_database(database_name)

    def find_record(self, collection_name, query):
        """
        Return the first record mathing the given Mongo query.

        :param collection_name: Name of the collection to search in.
        :param query: MongoDB Query, e.g. {_id: 123}
        :return: A single MongoDB record or None if not found.

        :raise DataSourceError
        """
        cursor = self._get_collection(collection_name).find(query)
        for record in cursor:
            # Return the first record found.
            return record
        # Return None if nothing found.
        return None

    def find_records(self, collection_name, query={}, sort_by=None,
                     sort_direction=None, start=0, limit=None):
        """
        Return a cursor of records from the given MongoDB collection.

        :param collection_name: Name of the MongoDB collection to query.
        :param query: Standard MongoDB query. By default no restriction.
        :param sort_by: Name of a single field to sort by.
        :param sort_direction: The direction to sort, "asc" or "desc".
        :param start: Skip first n results.
        :param limit: The maximum number of results to return.

        :return: Cursor -- An iterable with results.
        :raise DataSourceError
        """
        cursor = self._get_collection(collection_name).find(query)
        if sort_by is not None:
            cursor = self._apply_sort(cursor, sort_by, sort_direction)
        cursor = cursor.skip(start)
        if limit is not None:
            cursor = cursor.limit(limit)
        return MongoDbCursor(cursor)

    def _get_database(self, database_name):
        """
        Get PyMongo client pointing to the current database.

        :return: MongoDB client of the current database.
        :raise DataSourceError
        """
        try:
            return self._client[database_name]
        except InvalidName as ex:
            raise DataSourceError("Cannot connect to database %s!"
                                  % self._database) from ex

    def _get_collection(self, collection_name):
        """
        Get PyMongo client pointing to the current DB and the given collection.

        :return: MongoDB client of the current database and given collection.
        :raise DataSourceError
        """
        try:
            return self._database[collection_name]
        except InvalidName as ex:
            raise DataSourceError("Cannot access MongoDB collection %s!"
                                  % collection_name) from ex
        except Exception as ex:
            raise DataSourceError("Unexpected error when accessing MongoDB"
                                  "collection %s!"
                                  % collection_name) from ex

    def _apply_sort(self, cursor, sort_by, sort_direction):
        """
        Apply sort to a cursor.

        :param cursor: The cursor to apply sort on.
        :param sort_by: The field name to sort by.
        :param sort_direction: The direction to sort, "asc" or "desc".
        :return:

        """
        if sort_direction is not None and sort_direction.lower() == "desc":
            sort = pymongo.DESCENDING
        else:
            sort = pymongo.ASCENDING
        return cursor.sort(sort_by, sort)



# coding=utf-8
"""WebAPI module for handling run-related requests."""
import json

from flask import current_app, request, Response, render_template


def parse_int_arg(name, default):
    """Return a given URL parameter as int or return the default value."""
    return default if request.args.get(name) is None \
        else int(request.args.get(name))


def parse_query_filter():
    """Parse the Run query filter from the URL as a dictionary."""
    query_string = request.args.get("queryFilter")
    if query_string is None:
        return {"type": "and", "filters": []}
    query = json.loads(query_string)
    assert type(query) == dict
    assert type(query.get("type")) == str
    return query


def get_runs():
    """Get all runs, sort it and return a response."""
    data = current_app.config["data"]
    draw = parse_int_arg("draw", 1)
    start = parse_int_arg("start", 0)
    length = parse_int_arg("length", -1)
    length = length if length >= 0 else None
    order_column = request.args.get("order[0][column]")
    order_dir = request.args.get("order[0][dir]")

    query = parse_query_filter()

    if order_column is not None:
        order_column = \
            request.args.get("columns[%d][name]" % int(order_column))
        if order_column == "hostname":
            order_column = "host.hostname"

    runs = data.get_runs(
        start=start, limit=length,
        sort_by=order_column, sort_direction=order_dir, query=query)
    # records_total should be the total size of the records in the database,
    # not what was returned
    records_total = runs.count()
    records_filtered = runs.count()
    return Response(render_template(
        "api/runs.js", runs=runs,
        draw=draw, recordsTotal=records_total,
        recordsFiltered=records_filtered),
        mimetype="application/json")



"""Errors that might occur during data access."""


class NotFoundError(Exception):
    """Record not found exception."""

    def __init__(self, *args, **kwargs):
        """Record not found exception."""
        Exception.__init__(self, *args, **kwargs)


class DataSourceError(Exception):
    """Error when accessing the data source."""

    def __init__(self, *args, **kwargs):
        """Error when accessing the data source."""
        Exception.__init__(self, *args, **kwargs)



# coding=utf-8
"""Module for launching processes and reading their output."""
import atexit
import os
import select
import time
from subprocess import Popen, PIPE


class Process:
    """A process that can be run and read output from."""

    instances = []  # type: List[Process]
    """Instances of all processes."""

    def __init__(self, command):
        """
        Define a new process but do not start it.

        :param command: A command to start. Parameters separated with spaces
         or as a list, e.g. "command arg1 arg2" or ["command", "arg1", "arg2"].
        """
        if type(command == list):
            self.command = command
        else:
            self.command = command.split(" ")

        self.proc = None  # type: Popen

    def run(self):
        """Run the process."""
        environment = os.environ.copy()
        # necessary for reading from processes that don't flush
        # stdout automatically
        environment["PYTHONUNBUFFERED"] = "1"
        self.proc = Popen(self.command, env=environment, stdout=PIPE)
        Process.instances.append(self)

    def is_running(self):
        """Test if the process is running."""
        return self.proc is not None and not bool(self.proc.poll())

    def read_line(self, time_limit=None):
        """
        Read a line from the process.

        Block or wait for time_limit secs. Timeout does not work on Windows.
        """
        if self.proc is not None:
            poll_obj = select.poll()
            poll_obj.register(self.proc.stdout, select.POLLIN)
            start = time.time()
            while time_limit is None or time.time() - start < time_limit:
                poll_result = poll_obj.poll(0)
                if poll_result:
                    line = self.proc.stdout.readline().decode()
                    return line
                else:
                    time.sleep(0.05)
            raise TimeoutError()
        else:
            return None

    def terminate(self, wait=False):
        """Terminate the process."""
        if self.proc is not None:
            self.proc.stdout.close()
            self.proc.terminate()
            if wait:
                self.proc.wait()

    def pid(self):
        """Get the process id. Returns none for non-running processes."""
        if self.proc is not None:
            return self.proc.pid
        else:
            return None

    @staticmethod
    def terminate_all(wait=False):
        """
        Terminate all processes.

        :param wait: Wait for each to terminate
        :type wait: bool
        :return:
        :rtype:
        """
        for instance in Process.instances:
            if instance.is_running():
                instance.terminate(wait)

    @staticmethod
    def create_process(command):
        """
        Create a process using this factory method. This does not start it.

        :param command: A command to start. Parameters separated with spaces
         or as a list, e.g. "command arg1 arg2" or ["command", "arg1", "arg2"].
        """
        if getattr(select, "poll", None) is not None:
            return Process(command)
        else:
            return WindowsProcess(command)


class WindowsProcess(Process):
    """A class for a Windows process."""

    def __init__(self, command):
        """
       Define a new process but do not start it.

       :param command: A command to start. Parameters separated with spaces
        or as a list, e.g. "command arg1 arg2" or ["command", "arg1", "arg2"].
       """
        Process.__init__(self, command)

    def read_line(self, time_limit=None):
        """
        Read a line from the process.

        On Windows, this the time_limit has no effect, it always blocks.
        """
        if self.proc is not None:
            return self.proc.stdout.readline().decode()
        else:
            return None


# Kill all on exit
atexit.register(Process.terminate_all)


class ProcessError(Exception):
    """A process-related exception."""

    pass


class UnexpectedOutputError(ProcessError):
    """An unexpected output produced by the process."""

    def __init__(self, output, expected=None):
        """Create an unexpected output exception."""
        self.expected = expected
        self.output = output



"""Module for managing TensorBoard processes."""
import re

from sacredboard.app.process.process \
    import Process, ProcessError, UnexpectedOutputError

TENSORBOARD_BINARY = "tensorboard"


def stop_all_tensorboards():
    """Terminate all TensorBoard instances."""
    for process in Process.instances:
        print("Process '%s', running %d" % (process.command[0],
                                            process.is_running()))
        if process.is_running() and process.command[0] == "tensorboard":
            process.terminate()


class TensorboardNotFoundError(ProcessError):
    """TensorBoard binary not found."""

    pass


def run_tensorboard(logdir, listen_on="0.0.0.0", tensorboard_args=None):
    """
    Launch a new TensorBoard instance.

    :param logdir: Path to a TensorFlow summary directory
    :param listen_on: The IP address TensorBoard should listen on.
    :param tensorboard_args: Additional TensorBoard arguments.
    :return: Returns the port TensorBoard is listening on.
    :raise UnexpectedOutputError
    :raise TensorboardNotFoundError
    """
    if tensorboard_args is None:
        tensorboard_args = []
    tensorboard_instance = Process.create_process(
        TENSORBOARD_BINARY.split(" ") +
        ["--logdir", logdir, "--host", listen_on] + tensorboard_args)
    try:
        tensorboard_instance.run()
    except FileNotFoundError as ex:
        raise TensorboardNotFoundError(ex)

    # Read first line of output from tensorboard - it should contain
    # the port where it is listening on
    data = tensorboard_instance.read_line(time_limit=10)
    search = re.search("on port ([0-9]+)", data)
    if search is not None:
        port = search.group(1)
        return port
    else:
        tensorboard_instance.terminate()
        raise UnexpectedOutputError(
            data,
            expected="The port that Tensorboard is listening on."
        )



"""
Implementation of cursor for iterating over results.

Backed by pymongo cursor.
"""
from sacredboard.app.data.datastorage import Cursor


class MongoDbCursor(Cursor):
    """Implements Cursor for mongodb."""

    def __init__(self, mongodb_cursor):
        """Initialize a MongoDB cursor."""
        self.mongodb_cursor = mongodb_cursor

    def count(self):
        """Return the number of items in this cursor."""
        return self.mongodb_cursor.count()

    def __iter__(self):
        """Iterate over runs."""
        return self.mongodb_cursor



"""Implements backend storage interface for sacred's file store."""

import datetime
import os
import json

from sacredboard.app.data.datastorage import Cursor, DataStorage

CONFIG_JSON = "config.json"
RUN_JSON = "run.json"
INFO_JSON = "info.json"


def _path_to_file(basepath, run_id, file_name):
    return os.path.join(basepath, str(run_id), file_name)


def _path_to_config(basepath, run_id):
    return _path_to_file(basepath, str(run_id), CONFIG_JSON)


def _path_to_info(basepath, run_id):
    return _path_to_file(basepath, str(run_id), INFO_JSON)


def _path_to_run(basepath, run_id):
    return os.path.join(basepath, str(run_id), RUN_JSON)


def _read_json(path_to_json):
    with open(path_to_json) as f:
        return json.load(f)


def _create_run(run_id, runjson, configjson, infojson):
    runjson["_id"] = run_id
    runjson["config"] = configjson
    runjson["info"] = infojson

    # TODO probably want a smarter way of detecting
    # which values have type "time."
    for k in ["start_time", "stop_time", "heartbeat"]:
        runjson[k] = datetime.datetime.strptime(runjson[k],
                                                '%Y-%m-%dT%H:%M:%S.%f')
    return runjson


class FileStoreCursor(Cursor):
    """Implements the cursor for file stores."""

    def __init__(self, count, iterable):
        """Initialize FileStoreCursor with a given iterable."""
        self.iterable = iterable
        self._count = count

    def count(self):
        """
        Return the number of runs in this query.

        :return: int
        """
        return self._count

    def __iter__(self):
        """Iterate over runs."""
        return iter(self.iterable)


class FileStorage(DataStorage):
    """Object to interface with one of sacred's file stores."""

    def __init__(self, path_to_dir):
        """Initialize file storage run accessor."""
        super().__init__()
        self.path_to_dir = os.path.expanduser(path_to_dir)

    def get_run(self, run_id):
        """
        Return the run associated with a particular `run_id`.

        :param run_id:
        :return: dict
        :raises FileNotFoundError
        """
        config = _read_json(_path_to_config(self.path_to_dir, run_id))
        run = _read_json(_path_to_run(self.path_to_dir, run_id))
        info = _read_json(_path_to_info(self.path_to_dir, run_id))
        return _create_run(run_id, run, config, info)

    def get_runs(self, sort_by=None, sort_direction=None,
                 start=0, limit=None, query={"type": "and", "filters": []}):
        """
        Return all runs in the file store.

        If a run is corrupt, e.g. missing files, it is skipped.

        :param sort_by: NotImplemented
        :param sort_direction:  NotImplemented
        :param start: NotImplemented
        :param limit: NotImplemented
        :param query: NotImplemented
        :return: FileStoreCursor
        """
        all_run_ids = os.listdir(self.path_to_dir)

        def run_iterator():
            blacklist = set(["_sources"])
            for id in all_run_ids:
                if id in blacklist:
                    continue
                try:
                    yield self.get_run(id)
                except FileNotFoundError:
                    # An incomplete experiment is a corrupt experiment.
                    # Skip it for now.
                    # TODO
                    pass

        count = len(all_run_ids)
        return FileStoreCursor(count, run_iterator())



"""
Module responsible for accessing the Metrics data in MongoDB.

Issue: https://github.com/chovanecm/sacredboard/issues/60
"""

from bson import ObjectId
from bson.errors import InvalidId

from sacredboard.app.data import NotFoundError
from .genericdao import GenericDAO
from ..metricsdao import MetricsDAO


class MongoMetricsDAO(MetricsDAO):
    """Implementation of MetricsDAO for MongoDB."""

    def __init__(self, generic_dao: GenericDAO):
        """
        Create new metrics accessor for MongoDB.

        :param generic_dao: A configured generic MongoDB data access object
         pointing to an appropriate database.
        """
        self.generic_dao = generic_dao
        self.metrics_collection_name = "metrics"
        """Name of the MongoDB collection with metrics."""

    def get_metric(self, run_id, metric_id):
        """
        Read a metric of the given id and run.

        The returned object has the following format (timestamps are datetime
         objects).

        .. code::

            {"steps": [0,1,20,40,...],
            "timestamps": [timestamp1,timestamp2,timestamp3,...],
            "values": [0,1 2,3,4,5,6,...],
            "name": "name of the metric",
            "metric_id": "metric_id",
            "run_id": "run_id"}

        :param run_id: ID of the Run that the metric belongs to.
        :param metric_id: The ID fo the metric.
        :return: The whole metric as specified.

        :raise NotFoundError
        """
        query = self._build_query(run_id, metric_id)
        row = self._read_metric_from_db(metric_id, run_id, query)
        metric = self._to_intermediary_object(row)
        return metric

    def _read_metric_from_db(self, metric_id, run_id, query):
        row = self.generic_dao.find_record(self.metrics_collection_name, query)
        if row is None:
            raise NotFoundError("Metric %s for run %s not found."
                                % (metric_id, run_id))
        return row

    def _build_query(self, run_id, metric_id):
        # Metrics in MongoDB is always an ObjectId
        try:
            id = ObjectId(metric_id)
            return {"run_id": run_id, "_id": id}
        except InvalidId as ex:
            raise NotFoundError("Metric Id %s is invalid "
                                "ObjectId in MongoDB" % metric_id) from ex

    def _to_intermediary_object(self, row):
        return {
            "metric_id": str(row["_id"]),
            "run_id": row["run_id"],
            "name": row["name"],
            "steps": row["steps"],
            "timestamps": row["timestamps"],
            "values": row["values"],
        }



"""
Accessing the Metrics Data.

Issue: https://github.com/chovanecm/sacredboard/issues/58
"""
import datetime
from flask import Response, render_template, Blueprint, current_app

from sacredboard.app.data import NotFoundError

metrics = Blueprint("metrics", __name__)


def get_metric(run_id, metric_id):
    """
    Get a specific Sacred metric from the database.

    Returns a JSON response or HTTP 404 if not found.
    Issue: https://github.com/chovanecm/sacredboard/issues/58
    """
    data = current_app.config["data"]  # type: DataStorage
    dao = data.get_metrics_dao()
    metric = dao.get_metric(run_id, metric_id)

    return Response(render_template(
        "api/metric.js",
        run_id=metric["run_id"],
        metric_id=metric["metric_id"],
        name=metric["name"],
        steps=metric["steps"],
        timestamps=metric["timestamps"],
        values=metric["values"]),
        mimetype="application/json")


@metrics.app_template_filter("format_metric_timestamp")
def format_metric_timestamp(timestamp: datetime.datetime):
    """
    Convert given timestamp to UTC string format.

    :param timestamp UTC time to convert.
    :type timestamp datetime.datetime

    :return str
    """
    return "%sZ" % timestamp.isoformat()


@metrics.route("/api/run/<int:run_id>/metric/<metric_id>")
def api_metric(run_id, metric_id):
    """
    Get Sacred metrics.

    Issue: https://github.com/chovanecm/sacredboard/issues/58
    The API will return an object of the metric with parameters

    {steps: [0,1,20,40,...],
    strtimesteps: [timestep1,timestep2,timestep3,...],
     values: [0,1 2,3,4,5,6,...],
     name: "name of the metric"}

    :param run_id: The ID of the run to search the metric for.
    :param metric_id: The metric ID.
    :return: JSON object in the given format.
    """
    return get_metric(run_id, metric_id)


@metrics.errorhandler(NotFoundError)
def handle_not_found_error(e):
    """Handle exception when a metric is not found."""
    return "Couldn't find resource:\n%s" % e, 404


def initialize(app):
    """Register the module in Flask."""
    app.register_blueprint(metrics)



"""Interfaces for data storage."""
from .errors import NotFoundError
from .metricsdao import MetricsDAO


class Cursor:
    """Interface that abstracts the cursor object returned from databases."""

    def __init__(self):
        """Declare a new cursor to iterate over runs."""
        pass

    def count(self):
        """Return the number of items in this cursor."""
        raise NotImplemented()

    def __iter__(self):
        """Iterate over elements."""
        raise NotImplemented()


class DataStorage:
    """
    Interface for data backends.

    Defines the API for various data stores
    databases, file stores, etc. --- that sacred supports.
    """

    def __init__(self):
        """Initialize data accessor."""
        pass

    def get_run(self, run_id):
        """Return the run associated with the id."""
        raise NotImplemented()

    def get_runs(self, sort_by=None, sort_direction=None,
                 start=0, limit=None, query={"type": "and", "filters": []}):
        """Return all runs that match the query."""
        raise NotImplemented()

    def get_metrics_dao(self):
        """
        Return a data access object for metrics.

        By default, returns a dummy Data Access Object if not overridden.
        Issue: https://github.com/chovanecm/sacredboard/issues/62

        :return MetricsDAO
        """
        return DummyMetricsDAO()


class DummyMetricsDAO(MetricsDAO):
    """Dummy Metrics DAO that does not find any metric."""

    def get_metric(self, run_id, metric_id):
        """
        Raise NotFoundError. Always.

        :raise NotFoundError
        """
        raise NotFoundError("Metrics not supported by this backend.")



"""Sacred(board) Data Access Layer."""
from .datastorage import Cursor, DataStorage
from sacredboard.app.data.errors import NotFoundError, DataSourceError
from .metricsdao import MetricsDAO

__all__ = ["Cursor", "DataStorage", "MetricsDAO", "NotFoundError",
           "DataSourceError"]



# coding=utf-8
"""Accesses data in Sacred's MongoDB."""
import bson
import pymongo

from sacredboard.app.data.datastorage import Cursor, DataStorage
from sacredboard.app.data.pymongo import GenericDAO, MongoMetricsDAO


class MongoDbCursor(Cursor):
    """Implements Cursor for mongodb."""

    def __init__(self, mongodb_cursor):
        """Initialize a MongoDB cursor."""
        self.mongodb_cursor = mongodb_cursor

    def count(self):
        """Return the number of items in this cursor."""
        return self.mongodb_cursor.count()

    def __iter__(self):
        """Iterate over runs."""
        return self.mongodb_cursor


class PyMongoDataAccess(DataStorage):
    """Access records in MongoDB."""

    RUNNING_DEAD_RUN_CLAUSE = {
        "status": "RUNNING", "$where": "new Date() - this.heartbeat > 120000"}
    RUNNING_NOT_DEAD_CLAUSE = {
        "status": "RUNNING", "$where": "new Date() - this.heartbeat <= 120000"}

    def __init__(self, uri, database_name, collection_name):
        """
        Set up MongoDB access layer, don't connect yet.

        Better use the static methods build_data_access
        or build_data_access_with_uri
        """
        super().__init__()
        self._uri = uri
        self._db_name = database_name
        self._client = None
        self._db = None
        self._collection_name = collection_name
        self._generic_dao = None

    def connect(self):
        """Initialize the database connection."""
        self._client = self._create_client()
        self._db = getattr(self._client, self._db_name)
        self._generic_dao = GenericDAO(self._client, self._db_name)

    def _create_client(self):
        """Return a new Mongo Client."""
        return pymongo.MongoClient(host=self._uri)

    def get_runs(self, sort_by=None, sort_direction=None,
                 start=0, limit=None, query={"type": "and", "filters": []}):
        """
        Return multiple runs (default all).

        The query format (optional):

        {"type": "and", "filters": [
         {"field": "host.hostname", "operator": "==", "value": "ntbacer"},
         {"type": "or", "filters": [
            {"field": "result", "operator": "==", "value": 2403.52},
            {"field": "host.python_version", "operator": "==", "value":"3.5.2"}
            ]}]}

        The parameter is built from clauses.
        Each clause is either conjunctive (``"and"``), disjunctive (``"or"``),
        or a *terminal clause*. Each of the the earlier two types must specify
        the ``"filters`` array of other clauses to be joined together
        by that logical connective (and/or).

        A terminal clause does not specifies its type, instead,
        it has a different set of fields:
        the ``field`` to be queried on (based on the MongoDB schema,
        using dot notation to access nested documents), the ``operator``
        (one of ``"=="``, ``"!="``, ``"<"``, ``"<="``, ``">``, ``">="``,
        and ``"regex"`` for regular expressions).
        The ``value`` field contains the value to be compared with
        (either a string or a number). Notice that for the ``status`` field,
        the ``RUNNING`` and ``DEAD`` runs are compared by
        :func:`~PyMongoDataAccess.RUNNING_DEAD_RUN_CLAUSE` and
        :func:`~PyMongoDataAccess.RUNNING_NOT_DEAD_CLAUSE`
        """
        mongo_query = self._to_mongo_query(query)
        cursor = getattr(self._db, self._collection_name).find(mongo_query)
        if sort_by is not None:
            cursor = self._apply_sort(cursor, sort_by, sort_direction)
        cursor = cursor.skip(start)
        if limit is not None:
            cursor = cursor.limit(limit)

        return MongoDbCursor(cursor)

    def get_run(self, run_id):
        """
        Get a single run from the database.

        :param run_id: The ID of the run.
        :return: The whole object from the database.
        """
        try:
            cursor = getattr(self._db, self._collection_name) \
                .find({"_id": int(run_id)})
        except ValueError:
            # Probably not a number.
            cursor = getattr(self._db, self._collection_name) \
                .find({"_id": bson.ObjectId(run_id)})
        run = None
        for c in cursor:
            run = c
        return run

    @staticmethod
    def _apply_sort(cursor, sort_by, sort_direction):
        """
        Apply sort to a cursor.

        :param cursor: The cursor to apply sort on.
        :param sort_by: The field name to sort by.
        :param sort_direction: The direction to sort, "asc" or "desc".
        :return:

        """
        if sort_direction is not None and sort_direction.lower() == "desc":
            sort = pymongo.DESCENDING
        else:
            sort = pymongo.ASCENDING
        return cursor.sort(sort_by, sort)

    @staticmethod
    def _to_mongo_query(query):
        """
        Convert the query received by the Sacred Web API to a MongoDB query.

        Takes a query in format
        {"type": "and", "filters": [
         {"field": "host.hostname", "operator": "==", "value": "ntbacer"},
         {"type": "or", "filters": [
            {"field": "result", "operator": "==", "value": 2403.52},
            {"field": "host.python_version", "operator": "==", "value":"3.5.2"}
            ]}]}
        and returns an appropriate MongoDB Query.
        :param query: A query in the Sacred Web API format.
        :return: Mongo Query.
        """
        mongo_query = []
        for clause in query["filters"]:
            if clause.get("type") is None:
                mongo_clause = PyMongoDataAccess. \
                    _simple_clause_to_query(clause)
            else:
                # It's a subclause
                mongo_clause = PyMongoDataAccess._to_mongo_query(clause)
            mongo_query.append(mongo_clause)

        if len(mongo_query) == 0:
            return {}
        if query["type"] == "and":
            return {"$and": mongo_query}
        elif query["type"] == "or":
            return {"$or": mongo_query}
        else:
            raise ValueError("Unexpected query type %s" % query.get("type"))

    @staticmethod
    def _simple_clause_to_query(clause):
        """
        Convert a clause from the Sacred Web API format to the MongoDB format.

        :param clause: A clause to be converted. It must have "field",
        "operator" and "value" fields.
        :return: A MongoDB clause.
        """
        # It's a regular clause
        mongo_clause = {}
        value = clause["value"]
        if clause["field"] == "status" and clause["value"] in ["DEAD",
                                                               "RUNNING"]:
            return PyMongoDataAccess. \
                _status_filter_to_query(clause)
        if clause["operator"] == "==":
            mongo_clause[clause["field"]] = value
        elif clause["operator"] == ">":
            mongo_clause[clause["field"]] = {"$gt": value}
        elif clause["operator"] == ">=":
            mongo_clause[clause["field"]] = {"$gte": value}
        elif clause["operator"] == "<":
            mongo_clause[clause["field"]] = {"$lt": value}
        elif clause["operator"] == "<=":
            mongo_clause[clause["field"]] = {"$lte": value}
        elif clause["operator"] == "!=":
            mongo_clause[clause["field"]] = {"$ne": value}
        elif clause["operator"] == "regex":
            mongo_clause[clause["field"]] = {"$regex": value}
        return mongo_clause

    @staticmethod
    def _status_filter_to_query(clause):
        """
        Convert a clause querying for an experiment state RUNNING or DEAD.

        Queries that check for experiment state RUNNING and DEAD need to be
        replaced by the logic that decides these two states as both of them
        are stored in the Mongo Database as "RUNNING". We use querying by last
        heartbeat time.

        :param clause: A clause whose field is "status" and "value" is one of
        RUNNING, DEAD.
        :return: A MongoDB clause.
        """
        if clause["value"] == "RUNNING":
            mongo_clause = PyMongoDataAccess.RUNNING_NOT_DEAD_CLAUSE
        elif clause["value"] == "DEAD":
            mongo_clause = PyMongoDataAccess.RUNNING_DEAD_RUN_CLAUSE
        if clause["operator"] == "!=":
            mongo_clause = {"$not": mongo_clause}
        return mongo_clause

    @staticmethod
    def build_data_access(host, port, database_name, collection_name):
        """
        Create data access gateway.

        :param host: The database server to connect to.
        :type host: str
        :param port: Database port.
        :type port: int
        :param database_name: Database name.
        :type database_name: str
        :param collection_name: Name of the collection with Sacred runs.
        :type collection_name: str
        """
        return PyMongoDataAccess("mongodb://%s:%d" % (host, port),
                                 database_name, collection_name)

    @staticmethod
    def build_data_access_with_uri(uri, database_name, collection_name):
        """
        Create data access gateway given a MongoDB URI.

        :param uri: Connection string as defined in
                https://docs.mongodb.com/manual/reference/connection-string/
        :type uri: str
        :param database_name: Database name
        :type database_name: str
        :param collection_name: Name of the collection
        where Sacred stores its runs
        :type collection_name: str
        """
        return PyMongoDataAccess(uri, database_name, collection_name)

    def get_metrics_dao(self):
        """
        Return a data access object for metrics.

        The method can be called only after a connection to DB is established.
        Issue: https://github.com/chovanecm/sacredboard/issues/62

        :return MetricsDAO
        """
        return MongoMetricsDAO(self._generic_dao)



from setuptools import setup, find_packages

with open('description.txt') as f:
    long_description = ''.join(f.readlines())


def get_requirements():
    with open("requirements.txt") as f:
        return f.readlines()


setup(
    author="Martin Chovanec",
    author_email="chovamar@fit.cvut.cz",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
    ],
    description="Sacredboard - accessing Sacred data",
    long_description=long_description,
    license="MIT License",
    url="https://github.com/chovanecm/sacredboard",
    name="sacredboard",
    keywords="sacred",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "sacredboard = sacredboard.bootstrap:run"
        ]
    },
    install_requires=get_requirements(),
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "mongomock"],
    version="0.3"
)



# coding=utf-8
"""Define HTTP endpoints for the Sacredboard Web API."""
import re
from pathlib import Path

from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request, Response, redirect, url_for

from ..process.tensorboard import TensorboardNotFoundError,\
                                  stop_all_tensorboards
from ..process import process, tensorboard
from ..webapi.runs import get_runs

routes = Blueprint("routes", __name__)


@routes.route("/")
def index():
    """Redirect user to the main page."""
    return redirect(url_for("routes.show_runs"))


@routes.route("/_tests")
def tests():
    """Redirect user to a page with JavaScript tests."""
    return redirect(url_for("static", filename="scripts/tests/index.html"))


@routes.route("/runs")
def show_runs():
    """Render the main page with a list of experiment runs."""
    # return render_template("runs.html", runs=data.runs(), type=type)
    return render_template("runs.html", runs=[], type=type)


@routes.route("/api/run")
def api_runs():
    """Return a list of runs as a JSON object."""
    return get_runs()


@routes.route("/api/run/<run_id>")
def api_run(run_id):
    """Return a single run as a JSON object."""
    data = current_app.config["data"]
    run = data.get_run(run_id)
    records_total = 1 if run is not None else 0
    if records_total == 0:
        return Response(
            render_template(
                "api/error.js",
                error_code=404,
                error_message="Run %s not found." % run_id),
            status=404,
            mimetype="application/json")
    records_filtered = records_total
    return Response(render_template("api/runs.js", runs=[run], draw=1,
                                    recordsTotal=records_total,
                                    recordsFiltered=records_filtered,
                                    full_object=True),
                    mimetype="application/json")


@routes.route("/tensorboard/start/<run_id>/<int:tflog_id>")
def run_tensorboard(run_id, tflog_id):
    """Launch TensorBoard for a given run ID and log ID of that run."""
    data = current_app.config["data"]
    # optimisticaly suppose the run exists...
    run = data.get_run(run_id)
    base_dir = Path(run["experiment"]["base_dir"])
    log_dir = Path(run["info"]["tensorflow"]["logdirs"][tflog_id])
    # TODO ugly!!!
    if log_dir.is_absolute():
        path_to_log_dir = log_dir
    else:
        path_to_log_dir = base_dir.joinpath(log_dir)

    port = int(tensorboard.run_tensorboard(str(path_to_log_dir)))
    url_root = request.url_root
    url_parts = re.search("://([^:/]+)", url_root)
    redirect_to_address = url_parts.group(1)
    return redirect("http://%s:%d" % (redirect_to_address, port))


@routes.route("/tensorboard/stop", methods=['GET', 'POST'])
def close_tensorboards():
    """Stop all TensorBoard instances launched by Sacredboard."""
    stop_all_tensorboards()
    return "Stopping tensorboard"


@routes.errorhandler(TensorboardNotFoundError)
def handle_tensorboard_not_found(e):
    """Handle exception: tensorboard script not found."""
    return "Tensorboard not found on your system." \
           " Please install tensorflow first. Sorry.", 503


@routes.errorhandler(TimeoutError)
def handle_tensorboard_timeout(e):
    """Handle exception: TensorBoard does not respond."""
    return "Tensorboard does not respond. Sorry.", 503


@routes.errorhandler(process.UnexpectedOutputError)
def handle_tensorboard_unexpected_output(e: process.UnexpectedOutputError):
    """Handle Exception: TensorBoard has produced an unexpected output."""
    return "Tensorboard outputted '%s'," \
           " but the information expected was: '%s'. Sorry." \
           % (e.output, e.expected), 503


def setup_routes(app):
    """Register all HTTP endpoints defined in this file."""
    app.register_blueprint(routes)
